(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory DpTransImpl
imports TermImpl DpFramework Substitution
begin

type_synonym ('f, 'v) dp = "('f, 'v) rule" 
type_synonym ('f, 'v) dpL = "('f, 'v) dp list"

definition
  check_wf_trs :: "('f::show, 'v::show) rules \<Rightarrow> shows check"
where
  "check_wf_trs R \<equiv> (
     check_allm (\<lambda>rule. do {
       check (\<not> is_var (fst rule)) (shows_string ''variable left-hand side in'');
       check_subseteq (vars_term_list (snd rule)) (vars_term_list (fst rule))
         <+? (\<lambda>x. ''free variable '' +#+ shows x
               +@+ shows_string '' in right-hand side of'')
     } <+? (\<lambda>s. s +@+ '' rule '' +#+ shows_rule rule +@+ shows_nl)
     ) R
   <+? (\<lambda>e. ''the TRS is not well-formed'' +#+ shows_nl +@+ e)
   )"

lemma check_wf_trs_conf[simp]:
  shows "isOK (check_wf_trs R) = wf_trs (set R)"
proof
  assume "isOK (check_wf_trs R)"
  hence "\<forall>(l, r)\<in>set R. \<not> is_var l" unfolding check_wf_trs_def by auto
  hence "\<forall>(l, r)\<in>set R. \<exists>f ts. l = Fun f ts" by best
  moreover have "\<forall>(l, r)\<in>set R. vars_term r \<subseteq> vars_term l"
  proof (intro ballI2)
    fix l r assume "(l, r) \<in> set R"
    from `isOK (check_wf_trs R)`
      have "isOK(check_allm (\<lambda>(l, r). check_subseteq (vars_term_list r) (vars_term_list l)) R)"
      unfolding check_wf_trs_def by auto
    hence "\<forall>(l, r)\<in>set R. isOK (check_subseteq (vars_term_list r) (vars_term_list l))"
      unfolding split_def by simp
    with `(l, r) \<in> set R` show "vars_term r \<subseteq> vars_term l" by auto
  qed
  ultimately show "wf_trs (set R)" unfolding wf_trs_def by auto
next
  assume "wf_trs (set R)" thus "isOK (check_wf_trs R)"
    unfolding wf_trs_def check_wf_trs_def by force
qed

definition
  DP_list :: "('f \<Rightarrow> 'f) \<Rightarrow> ('f, 'v) rules \<Rightarrow> ('f,'v)rule_map \<Rightarrow> ('f, 'v) rules"
where
  "DP_list shp R RM \<equiv>
    let rm = the RM in        
    concat (map (\<lambda>lr.
      let l = fst lr; s = sharp_term shp l in
      map (\<lambda>u. (s, sharp_term shp u)) (filter (\<lambda>u. \<not> l \<rhd> u \<and> \<not> is_var u \<and> rm (the (root u), num_args u) \<noteq> []) (supteq_list (snd lr)))) R)"

lemma DP_list[simp]: assumes rm: "computeRuleMap R = Some rm"
  shows "set (DP_list shp R (computeRuleMap R)) = DP shp (set R)" (is "?L = ?R")
proof -
  from computeRuleMap_defs[OF rm] have rm_def: "\<And> f n. defined (set R) (f,n) = (rm (f,n) \<noteq> [])" unfolding defs_def by auto
  show ?thesis
  proof
    show "?L \<subseteq> ?R"
    proof
      fix s t assume "(s, t) \<in> ?L" 
      then obtain l r u where "(l, r) \<in> set R" and "r \<unrhd> u" and "\<not> l \<rhd> u"
        and "s = sharp_term shp l" and "t = sharp_term shp u"
        and "\<not> is_var u" and "defined (set R) (the (root u), num_args u)" unfolding rm_def unfolding DP_list_def Let_def by (auto simp: rm)
      thus "(s, t) \<in> DP shp (set R)" unfolding DP_def by (cases u) auto      
    qed
  next
    show "?R \<subseteq> ?L"
    proof
      fix s t assume "(s, t) \<in> ?R"
      then obtain lr u where "lr \<in> set R" and "snd lr \<unrhd> u" and "\<not> fst lr \<rhd> u"
        and "s = sharp_term shp (fst lr)" and "t = sharp_term shp u"
        and "\<not> is_var u" and "defined (set R) (the (root u), num_args u)" unfolding DP_def by auto
      thus "(s, t) \<in> ?L" unfolding DP_list_def Let_def rm_def by (auto simp: rm)
    qed
  qed
qed

definition
  check_DP :: "('f \<Rightarrow> 'f) \<Rightarrow> ('f::show,'v::show)rules \<Rightarrow> ('f,'v)rule_map \<Rightarrow> ('f,'v)dpL \<Rightarrow> shows check"
where
  "check_DP shp r rm P \<equiv> do {
     check_wf_trs r;
     let D = defs_list r;
     check_allm (\<lambda> (f,n). check ((shp f,n) \<notin> set D) (shows_string ''sharping '' +@+ shows f +@+ shows_string '' yields the defined symbol '' +@+ shows (shp f))) D;
     check_subseteq (DP_list shp r rm) P 
       <+? (\<lambda>dp. ''the DP '' +#+ shows_rule dp
         +@+ '' does not appear in the DP problem'' +#+ shows_nl)
   } <+? (\<lambda>s. ''the DP-transformation is not applied correctly.'' +#+ shows_nl +@+ s)"

lemma check_DP:
  assumes check: "isOK(check_DP shp r (computeRuleMap r) P)"
  shows "finite_dpp (set P, set r) \<Longrightarrow> SN (rstep (set r))"
proof -
  assume fin: "finite_dpp (set P, set r)"
  show "SN (rstep (set r))"
  proof -
    from check have wf: "wf_trs (set r)" 
      unfolding check_DP_def by auto
    with computeRuleMapSound2[of r] have "computeRuleMap r \<noteq> None" unfolding wf_trs_def by force
    then obtain rm where rm: "computeRuleMap r = Some rm" by (cases "computeRuleMap r", auto)
    from check have dp: "Trs.DP shp (set r) \<subseteq> set P"
      using DP_list[OF rm] unfolding check_DP_def Let_def by auto
    let ?DP = "Trs.DP shp (set r)"
    from fin have "finite_dpp (set P, set r)"  by simp
    hence fin2: "finite_dpp (set P \<inter> ?DP, set r)" by (unfold finite_dpp_def, auto)
    from dp have "?DP = set P \<inter> ?DP" by auto
    with fin2 have fin: "finite_dpp (DP shp (set r), set r)" by simp
    show "?thesis" 
    proof (rule finite_initial_dpp_imp_SN[OF wf _ fin])
      fix f n
      assume "defined (set r) (f,n)"
      thus "\<not> defined (set r) (shp f,n)"
        using check unfolding check_DP_def Let_def defs_list_sound by auto
    qed
  qed
qed

definition
  check_not_wf_trs :: "('f::show, 'v::show) rules \<Rightarrow> shows check"
where
  "check_not_wf_trs R \<equiv>
     check (\<not> isOK (check_wf_trs R)) (''The TRS is well formed'' +#+ shows_nl)"

lemma check_not_wf_trs:
  assumes "isOK(check_not_wf_trs R)"
  shows "\<not> SN (rstep (set R))"
proof -
  from assms have "\<not> wf_trs (set R)" unfolding check_not_wf_trs_def by auto
  with SN_rstep_imp_wf_trs show ?thesis by auto
qed

definition
  check_no_var :: "('f::show, 'v::show) term \<Rightarrow> shows check"
where
  "check_no_var t \<equiv> check (\<not> is_var t) (''variable found'' +#+ shows_nl)"

lemma check_no_var_sound[simp]:
  "isOK (check_no_var t) = (\<not> is_var t)"
using assms unfolding check_no_var_def by simp

definition
  check_no_defined_root :: "('f::show \<times> nat \<Rightarrow> ('f,'v)rules) \<Rightarrow> ('f, 'v::show) term \<Rightarrow> shows check"
where
  "check_no_defined_root rm t = check (rm (the (root t), num_args t) = []) (
     ''the root of '' +#+ shows t +@+ shows_string '' is defined''
   )"

lemma check_no_defined_root_sound:
  assumes "computeRuleMap R = Some rm" 
  shows "isOK (check_no_defined_root rm t) = (\<not> defined (set R) (the (root t), num_args t))"
using computeRuleMap_defs[OF assms] unfolding defs_def check_no_defined_root_def by auto 

lemma computeRuleMap_None_not_SN:
  assumes "computeRuleMap R = None"
  shows "\<not> SN_elt (rstep (set R)) t"
proof -
  from assms computeRuleMap_left_vars[of R] obtain x r where "(Var x,r) \<in> set R" by auto
  from left_var_imp_not_SN[OF this] show ?thesis .
qed
  
lemma [code]: "instance_rule lr st = (match_list [(fst lr, fst st), (snd lr, snd st)] \<noteq> None)" (is "?l = (match_list ?list \<noteq> None)")
proof
  assume ?l
  then obtain \<sigma> where fst: "fst lr = fst st \<cdot> \<sigma>" and snd: "snd lr = snd st \<cdot> \<sigma>" unfolding instance_rule_def by auto
  have "\<And> t l. (t, l) \<in> set ?list \<Longrightarrow> l \<cdot> \<sigma> = t" by (auto simp del: match_list.simps simp: fst snd)
  from match_list_complete[of ?list \<sigma>, OF this]
  obtain \<sigma>' where "match_list ?list = Some \<sigma>'" by blast
  thus "match_list ?list \<noteq> None" by simp
next
  assume "match_list ?list \<noteq> None"
  then obtain \<sigma> where "match_list ?list = Some \<sigma>" by (cases "match_list ?list", auto)
  from match_list_sound[OF this] have "fst lr = fst st \<cdot> \<sigma>" and "snd lr = snd st \<cdot> \<sigma>" by (force+)
  thus ?l unfolding instance_rule_def by auto
qed


definition check_CS_subseteq :: "('f,'v)rules \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f,'v)rule check"
where "check_CS_subseteq R S \<equiv> check_allm (\<lambda> (l,r). check (list_ex (instance_rule (l,r)) S) (l,r)) R"

lemma check_CS_subseteq[simp]: "isOK(check_CS_subseteq R S) = (CS (set R) \<subseteq> CS (set S))" (is "?l = ?r")
proof
  assume ?l
  show ?r
  proof
    fix x y
    assume "(x,y) \<in> CS (set R)"    
    thus "(x,y) \<in> CS (set S)"
    proof (induct)
      case (subst_closureI s t \<sigma>)
      with `?l`[unfolded check_CS_subseteq_def]
      obtain l r \<delta> where lr: "(l,r) \<in> set S" and s: "s = l \<cdot> \<delta>" and t: "t = r \<cdot> \<delta>"
        unfolding list_ex_iff by (force simp: instance_rule_def)
      show ?case unfolding s t
        using subst_closure.subst_closureI[OF lr, of "\<delta> \<circ>s \<sigma>"]
        by auto
    qed
  qed
next
  assume ?r
  show ?l
    unfolding check_CS_subseteq_def
  proof (simp, intro ballI)
    fix lr
    assume mem: "lr \<in> set R"
    obtain l r where lr: "lr = (l,r)" by (cases lr, auto)
    with mem have "(l,r) \<in> CS (set R)" using subset_subst_closure by auto
    with `?r` have "(l,r) \<in> CS (set S)" by auto
    thus "list_ex (instance_rule lr) S" unfolding lr
    proof (induct)
      case (subst_closureI s t \<sigma>)
      thus ?case unfolding list_ex_iff instance_rule_def by force
    qed
  qed
qed


end
