(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory InnermostImpl
imports "Q-Restricted_Rewriting" Substitution TermImpl
begin

definition
  NF_terms_list :: "('f, 'v) term list \<Rightarrow> ('f, 'v) term \<Rightarrow> bool"
where
  "NF_terms_list ls t = (list_all (\<lambda>s. list_all (\<lambda>l. \<not> matches s l) ls) (supteq_list t))"

lemma NF_terms_list_sound[simp]: "NF_terms_list Q t \<longleftrightarrow> t \<in> NF_terms (set Q)"
proof
  assume "NF_terms_list Q t"
  show "t \<in> NF_terms (set Q)"
  proof (rule ccontr)
    assume "t \<notin> NF_terms (set Q)"
    then obtain u where "(t, u) \<in> rstep (Id_on (set Q))" by auto
    then obtain C l r \<sigma> where "(l, r) \<in> Id_on (set Q)" and t: "t = C\<langle>l \<cdot> \<sigma>\<rangle>" by blast
    hence "t \<unrhd> l \<cdot> \<sigma>" by auto
    moreover from `NF_terms_list Q t`
      have "\<forall>s\<unlhd>t. \<forall>(l, r)\<in>Id_on (set Q). \<not> matches s l"
      unfolding NF_terms_list_def list_all_iff by auto
    ultimately show False using `(l, r) \<in> Id_on (set Q)` unfolding matches by force
  qed
next
  assume NF: "t \<in> NF_terms (set Q)"
  show "NF_terms_list Q t"
  proof (rule ccontr)
    assume "\<not> NF_terms_list Q t"
    then obtain s l \<sigma> where "t \<unrhd> s" and "l \<in> set Q" and s: "s = l \<cdot> \<sigma>"
      unfolding NF_terms_list_def list_all_iff matches by auto
    from `t \<unrhd> s` obtain C where t: "t = C\<langle>s\<rangle>" by auto
    from `l \<in> set Q` have "(l, l) \<in> Id_on (set Q)" by auto
    hence "(l \<cdot> \<sigma>, l \<cdot> \<sigma>) \<in> rstep (Id_on (set Q))" by auto
    hence "(t, t) \<in> rstep (Id_on (set Q))" unfolding s t by auto
    with NF show False by auto
  qed
qed

fun
  useful_rule_list :: "('f, 'v) term list \<Rightarrow> ('f, 'v) rule \<Rightarrow> bool"
where
  "useful_rule_list Q (Var _, _) = True" |
  "useful_rule_list Q (Fun f ss, _) = list_all (NF_terms_list Q) ss"

lemma useful_rule_list_sound[simp]: "useful_rule_list Q = useful_rule (set Q)"
proof (rule ext, unfold split_paired_all)
  fix l r show "useful_rule_list Q (l, r) = useful_rule (set Q) (l, r)"
  proof (cases l)
    case (Var x) show ?thesis unfolding Var useful_rule_def by auto
  next
    case (Fun f ts) show ?thesis
      by (simp add: NF_rstep_supt_args_conv useful_rule_def Fun list_all_iff)
  qed
qed

definition
  useful_rules_list :: "('f, 'v) term list \<Rightarrow> ('f, 'v) rule list  \<Rightarrow> ('f, 'v) rules"
where
  "useful_rules_list Q R = filter (useful_rule_list Q) R"

lemma useful_rules_list_sound[simp]:
  "set (useful_rules_list Q R) = useful_rules (set Q) (set R)"
unfolding useful_rules_def useful_rules_list_def by auto

end
