(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory RelativeRewriting
imports Trs RedPair
begin

type_synonym ('f,'v)reltrs = "('f,'v)term rel_ars"

type_synonym ('f,'v)reltrsL = "('f,'v)rules \<times> ('f,'v)rules"

fun rel_rstep :: "('f,'v)reltrs \<Rightarrow> ('f,'v)term ars"
where "rel_rstep (R,S) = rel_step (rstep R, rstep S)"

lemma one_step_ctxt_closed_rel_rstep : "one_step_ctxt_closed (rel_rstep RS)"
proof (cases RS)
  case (Pair R S)
  from rtrancl_one_step_ctxt_closed[OF one_step_ctxt_closed_rstep[of S]] have S: "one_step_ctxt_closed ((rstep S)^*)" .
  from Pair comp_one_step_ctxt_closed[OF comp_one_step_ctxt_closed[OF S one_step_ctxt_closed_rstep[of R] ] S]
  show ?thesis by (simp add: rel_step.simps)
qed
  
lemma subst_closed_rel_rstep : "subst_closed (rel_rstep RS)"
proof (cases RS)
  case (Pair R S)
  from subst_closed_rtrancl[OF subst_closed_rstep[of S]] have S: "subst_closed ((rstep S)^*)" .
  from Pair comp_subst_closed[OF S comp_subst_closed[OF subst_closed_rstep[of R] S] ]
  show ?thesis by (simp add: rel_step.simps)
qed
  

definition sound_rel_tt_impl :: "(('f,'v)reltrsL \<Rightarrow> ('f,'v)reltrsL \<Rightarrow> bool) \<Rightarrow> bool"
where "sound_rel_tt_impl tt \<equiv> \<forall> R S rR rS. (tt (R,S) (rR,rS) \<longrightarrow> rel_SN_alt (rstep (set rR),rstep (set rS)) \<longrightarrow> rel_SN_alt (rstep (set R),rstep (set S)))"

definition sound_rel_tt :: "(('f,'v)reltrs \<Rightarrow> ('f,'v)reltrs \<Rightarrow> bool) \<Rightarrow> bool"
where "sound_rel_tt tt \<equiv> \<forall> R S rR rS. (tt (R,S) (rR,rS) \<longrightarrow> rel_SN_alt (rstep rR, rstep rS) \<longrightarrow> rel_SN_alt (rstep R, rstep S))"

lemma apply_sound_rel_tt_impl : assumes sound: "sound_rel_tt_impl tt"
  and tt: "tt (R,S) (rR,rS)"
  and relSN: "rel_SN_alt (rstep (set rR), rstep (set rS))"
  shows "rel_SN_alt (rstep (set R), rstep (set S))"
using assms unfolding sound_rel_tt_impl_def by auto
      
  

context redpair
begin

lemma relativeRuleRemoval:
  assumes monotonic: "one_step_ctxt_closed \<succ>"
  and nonStrict: "r \<union> s \<subseteq> \<succeq> \<union> \<succ>"
  and strict: "(r - rr) \<union> (s - rs) \<subseteq> \<succ>"
  and rel_SN: "rel_SN_alt (rstep rr, rstep rs)"
  shows "rel_SN_alt (rstep r, rstep s)"
proof (simp only: rel_SN_alt.simps, intro allI impI)
  fix t
  assume allii: "\<forall>i. (t i, t (Suc i)) \<in> rstep r \<union> rstep s"
  hence alli: "\<forall>i. (t i, t (Suc i)) \<in> rstep (r \<union> s)" unfolding rstep_union .
  show "\<exists> i. \<forall> j \<ge> i. (t j, t (Suc j)) \<notin> rstep r"
  proof (cases "\<forall> i. \<exists> j \<ge> i. (t j, t (Suc j)) \<in> (rstep r - rstep rr) \<union> (rstep s - rstep rs)")
    case True
    from monotonic strict substS trs_inclusion have "(rstep (r - rr)) \<subseteq> \<succ>" by auto
    with rstep_diff have one: "rstep r - rstep rr \<subseteq> \<succ>" by auto
    from monotonic strict substS trs_inclusion have "(rstep (s - rs)) \<subseteq> \<succ>" by auto
    with rstep_diff have two: "rstep s - rstep rs \<subseteq> \<succ>" by auto
    from one two True alli have strict: "\<forall> i. \<exists> j \<ge> i. (t j, t (Suc j)) \<in> \<succ>" by blast 
    from trs_inclusion[OF union_one_step_ctxt_closed[OF wmNS monotonic] subst_closed_union[OF substNS substS] nonStrict]
      alli have "\<forall>i. (t i, t (Suc i)) \<in> (\<succeq> \<union> \<succ>)" by auto
    from non_strict_ending[OF this compat] SN obtain j where "\<forall> i \<ge> j. (t i, t (Suc i)) \<in> \<succeq> - \<succ>" unfolding SN_def by auto
    with strict have False by auto
    thus ?thesis ..
  next    
    case False
    then obtain i where not: "\<forall> j \<ge> i. (t j, t (Suc j)) \<notin> (rstep r - rstep rr) \<union> (rstep s - rstep rs)" by blast
    with allii have "\<forall> j \<ge> i. (t j, t (Suc j)) \<in> rstep (rr \<union> rs)" unfolding rstep_union by blast
    hence "\<forall> j. (shift t i j, shift t i (Suc j)) \<in> rstep (rr \<union> rs)" by auto
    with spec[OF rel_SN[simplified rel_SN_alt.simps], of "shift t i"] obtain j where not2: "\<forall> k \<ge> j. (shift t i k, shift t i (Suc k)) \<notin> rstep rr" by blast
    show ?thesis
    proof (rule exI[of _ "i + j"], intro allI impI)
      fix l
      assume l: "i + j \<le> l"
      then obtain k where k: "k = l - i" and kj: "k \<ge> j"  by auto
      from k l have shift: "(t l, t (Suc l)) = (shift t i k, shift t i (Suc k))" by simp
      from not l have not: "(t l, t (Suc l)) \<notin> (rstep r - rstep rr)" by auto
      from not2 kj have not2: "(t l, t (Suc l)) \<notin> rstep rr" by (simp add: shift)
      from not not2 show "(t l, t (Suc l)) \<notin> rstep r" by auto
    qed
  qed
qed
end

fun
  check_relative_rule_removal ::
    "('f :: show, 'v :: show)redpair_impl \<Rightarrow> 
     ('f,'v)reltrsL \<Rightarrow> ('f,'v)reltrsL \<Rightarrow> shows check"
where
  "check_relative_rule_removal rp (R,S) (rR,rS)  = do {
     let (Sr :: ('f,'v) rules , NSr) = List.partition (\<lambda> foo. isOK(rp_strict rp foo)) R;
     let (Ss :: ('f,'v) rules , NSs) = List.partition (\<lambda> foo. isOK(rp_strict rp foo)) S;
     rp_valid rp;
     rp_mono rp
       <+? (\<lambda>s. ''problem with monotonicity of strict order'' +#+ shows_nl +@+ s);
     check_allm (rp_non_strict rp) NSr
       <+? (\<lambda>s. ''problem when orienting rules'' +#+ shows_nl +@+ s);
     check_allm (rp_non_strict rp) NSs
       <+? (\<lambda>s. ''problem when orienting rules'' +#+ shows_nl +@+ s);
     check_subseteq NSs rS
       <+? (\<lambda>s. case (rp_strict rp) s of Inl err \<Rightarrow> err | Inr ok \<Rightarrow> internal_error);
     check_subseteq NSr rR
       <+? (\<lambda>s. case (rp_strict rp) s of Inl err \<Rightarrow> err | Inr ok \<Rightarrow> internal_error)
    } <+? (\<lambda>s. ''could not apply the rule removal processor with the following'' +#+ shows_nl +@+
         rp_desc rp +@+ shows_nl +@+ s)"

context generic_redpair_impl
begin

lemma relative_rule_removal_sound : 
  shows "sound_rel_tt_impl (\<lambda> RS rRS. isOK(check_relative_rule_removal (generate_redpair rp) RS rRS))"
unfolding sound_rel_tt_impl_def
proof (intro allI impI)
  let ?rp = "generate_redpair rp"
  let ?af = "rp_af ?rp"
  fix R S rR rS
  assume check: "isOK (check_relative_rule_removal ?rp (R,S) (rR,rS))"
    and relSN: "rel_SN_alt (rstep (set rR), rstep (set rS))"
  show "rel_SN_alt (rstep (set R), rstep (set S))"
  proof (cases "List.partition (\<lambda> r. isOK(rp_strict ?rp r)) R")
    case (Pair Sr NSr) note rPair = this
    let ?NSr = "set NSr \<inter> set rR"
    show ?thesis
    proof (cases "List.partition (\<lambda> r. isOK(rp_strict ?rp r)) S") 
      case (Pair Ss NSs) 
      let ?NSs = "set NSs \<inter> set rS"
      from Pair rPair check have 
        one: "isOK(check_allm (rp_strict ?rp) (Ss @ Sr))" and two: "isOK(check_allm (rp_non_strict ?rp) (NSs @ NSr))" 
        and mono: "isOK(rp_mono ?rp)" and valid: "isOK(rp_valid ?rp)"  by (auto simp: Let_def) 
      from generate_redpair[OF valid one two] mono obtain s NS where redpair: "mono_ce_af_redpair s NS ?af" and Sr: "set Sr \<subseteq> s" and NSr: "set NSr \<subseteq> NS" and Ss: "set Ss \<subseteq> s" and NSs: "set NSs \<subseteq> NS" and mono: "one_step_ctxt_closed s"
        by auto
      from rPair Sr NSr check have "set R - set NSr \<subseteq> s \<and> ?NSr \<inter> set R \<subseteq> NS \<union> s \<and> set NSr \<subseteq> set rR" by (auto)
      with check have NSr: "set R \<subseteq> NS \<union> s" and Sr: "set R - set rR \<subseteq> s" by auto
      from Pair Ss NSs check have "set S - set NSs \<subseteq> s \<and> ?NSs \<inter> set S \<subseteq> NS \<union> s \<and> set NSs \<subseteq> set rS" by auto
      with check have NSs: "set S \<subseteq> NS \<union> s" and Ss: "set S - set rS \<subseteq> s"  by auto
      from redpair interpret mono_ce_af_redpair s NS ?af .
      show ?thesis by (rule relativeRuleRemoval[OF mono _ _ relSN], auto simp: NSr NSs Sr Ss)
    qed
  qed
qed
end

lemmas rel_rstep_mono = rel_step_mono[OF rstep_mono rstep_mono, unfolded rel_rstep.simps[symmetric]]      
      
lemma rel_rstep_Id: "rel_rstep (R, S \<union> Id) = rel_rstep (R, S)"
proof -
  have "rstep (S^=) = (rstep S)^=" by (simp add: rstep_union rstep_id)
  thus ?thesis by (simp add: rel_step_Id)
qed

lemma rel_SN_map_imp_SN:
  assumes SN: "rel_SN_alt (rstep (map_trs fg R), rstep (map_trs fg S))"
  shows "rel_SN_alt (rstep R,rstep S)"
proof (unfold rel_SN_alt.simps, intro allI impI)
  fix f
  assume steps: "\<forall> i. (f i, f (Suc i)) \<in> rstep R \<union> rstep S"
  obtain g where g: "\<And> i. g i = map_term fg (f i)" by auto
  have "\<forall> i. (g i, g (Suc i)) \<in> rstep (map_trs fg R \<union> map_trs fg S)"
  proof
    fix i
    from spec[OF steps[simplified rstep_union], of i]
    show "(g i, g (Suc i)) \<in> rstep (map_trs fg R \<union> map_trs fg S)" unfolding g rstep_union
      using rstep_imp_map_rstep[where fg = fg] by blast
  qed
  with SN[unfolded rel_SN_alt.simps]
  obtain i where no: "\<forall> j \<ge> i. (g j, g (Suc j)) \<notin> rstep (map_trs fg R)" by blast
  show "\<exists> i. \<forall> j \<ge> i. (f j, f (Suc j)) \<notin> rstep R"
  proof (rule exI[of _ i], intro allI impI)
    fix j
    assume "i \<le> j"
    with no have "(g j, g (Suc j)) \<notin> rstep (map_trs fg R)" by simp
    with rstep_imp_map_rstep[where fg = fg] show "(f j, f (Suc j)) \<notin> rstep R" unfolding g by blast
  qed
qed

lemma rel_step_rtrancl: "rel_step (R,S^*) = rel_step (R,S)"
  unfolding rel_step.simps rtrancl_idemp by simp

end
