theory TermFun
imports TermFun_Setup
begin

text {* Below, we provide a rewriting view for some predefined concepts *}

text {* Register common types: *}

setup {*
  fold Term_Fun.register_datatypes
   (map single ["HOL.bool", "Nat.nat", "List.list", "Option.option"])
*}

text {*
  Assume globally that everything lives in the class.
  Constraints can be removed later...
*}

default_sort trs_rep

lemmas set_list_conv = 
  set.simps[symmetric] set_append[symmetric]
  List.append.simps

text {* Define TRSs for some standard functions. *}

define_trs plus_nat.simps
define_trs minus_nat.simps
define_trs take.simps
define_trs drop.simps
define_trs list.size(3-4)


fun le :: "nat \<Rightarrow> nat \<Rightarrow> bool"
where
  "le 0 n = True"
| "le (Suc n) 0 = False"
| "le (Suc n) (Suc m) = le n m"
define_trs le.simps
lemma le_subst[trs_subst]: "x \<le> y = (le x y)"
by (induct x y rule: le.induct) auto

fun eq where
  "eq 0 0 = True"
| "eq (Suc x) 0 = False"
| "eq 0 (Suc x) = False"
| "eq (Suc x) (Suc y) = eq x y"
define_trs eq.simps
lemma eq_subst[trs_subst]: "(x = y) = eq x y"
by (induct x y rule: eq.induct) auto


fun less where
  "less 0 0 = False"
| "less (Suc x) 0 = False"
| "less 0 (Suc x) = True"
| "less (Suc x) (Suc y) = less x y"
define_trs less.simps
lemma less_subst[trs_subst]: "(x < y) = less x y"
by (induct x y rule: less.induct) auto

end
