(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory TermImpl imports Term begin

fun supteq_list :: "('f,'v)term \<Rightarrow> ('f,'v)term list" 
where "supteq_list (Var x) = [Var x]"
  |   "supteq_list (Fun f ts) = Fun f ts # concat (map supteq_list ts)"


lemma supteq_list[simp]: "set(supteq_list t) = { s. t \<unrhd> s}"
proof (rule set_eqI, simp)
  fix s
  show "s \<in> set(supteq_list t) = (t \<unrhd> s)"
  proof (induct t, simp add: supteq_var_imp_eq)
    case (Fun f ss)
    show ?case
    proof (cases "Fun f ss = s", (auto)[1])
      case False
      show ?thesis
      proof
        assume "Fun f ss \<unrhd> s"
        with False have sup: "Fun f ss \<rhd> s" using supteq_supt_conv by auto
        obtain C where "C \<noteq> \<box>" and "Fun f ss = C\<langle>s\<rangle>" using sup by auto
        then obtain b D a where "Fun f ss = Fun f (b @ D\<langle>s\<rangle> # a)" by (cases C, auto)
        hence D: "D\<langle>s\<rangle> \<in> set ss" by auto
        with Fun[OF D] ctxt_imp_supteq[of D s] obtain t where "t \<in> set ss" and "s \<in> set (supteq_list t)" by auto
        thus "s \<in> set (supteq_list (Fun f ss))" by auto
      next
        assume "s \<in> set (supteq_list (Fun f ss))"
        with False obtain t where t: "t \<in> set ss" and "s \<in> set (supteq_list t)" by auto
        with Fun[OF t] have "t \<unrhd> s" by auto
        with t show "Fun f ss \<unrhd> s" by auto
      qed
    qed
  qed
qed

fun
  supt_impl :: "('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool"
where
  "supt_impl (Var x) t = False" |
  "supt_impl (Fun f ss) t = (t \<in> set ss \<or> (list_ex (\<lambda>s. supt_impl s t) ss))"

lemma supt_impl[code]: "op \<rhd> = (\<lambda>(x, y). supt_impl x y)"
proof (intro equalityI subsetI2)
  fix s t assume "s \<rhd> t" hence "supt_impl s t"
  proof (induct s)
    case (Var x) thus ?case by auto
  next
    case (Fun f ss) thus ?case
    proof (cases "t \<in> set ss")
      case True thus ?thesis by (simp)
    next
      case False
      assume "\<And>s. \<lbrakk>s \<in> set ss; s \<rhd> t\<rbrakk> \<Longrightarrow> supt_impl s t"
        and "Fun f ss \<rhd> t" and "t \<notin> set ss"
      moreover from `Fun f ss \<rhd> t` obtain s where "s \<in> set ss" and "s \<rhd> t"
        by (cases rule: supt.cases) (simp_all add: `t \<notin> set ss`)
      ultimately have "supt_impl s t" by simp
      with `s \<in> set ss` show ?thesis by (auto simp: list_ex_iff)
    qed
  qed
  thus "(s, t) \<in> (\<lambda>(x, y). supt_impl x y)" by (simp add: mem_def split_def)
next
  fix s t assume "(s, t) \<in> (\<lambda>(x, y). supt_impl x y)"
  hence "supt_impl s t" by (simp add: mem_def)
  thus "s \<rhd> t"
  proof (induct s)
    case (Var x) thus ?case by simp
  next
    case (Fun f ss)
    hence "t \<in> set ss \<or> (\<exists>s\<in>set ss. supt_impl s t)" by (simp add: list_ex_iff)
    thus ?case
    proof
      assume "t \<in> set ss" thus ?case by (auto intro: supt.intros)
    next
      assume "\<exists>s\<in>set ss. supt_impl s t"
      then obtain s where "s \<in> set ss" and "supt_impl s t" by auto
      with Fun have "s \<rhd> t" by auto
      with `s \<in> set ss` show ?thesis by (auto intro: supt.intros)
    qed
  qed
qed
lemma supteq_impl[code]: "op \<unrhd> = (\<lambda>(x, y). x = y \<or> supt_impl x y)"
unfolding supteq_supt_set_conv by (auto simp: supt_impl mem_def)

fun
  linear_term_impl :: "'v list \<Rightarrow> ('f, 'v) term \<Rightarrow> ('v list) option"
where
  "linear_term_impl xs (Var x) = (if x \<in> set xs then None else Some (x # xs))" |
  "linear_term_impl xs (Fun _ []) = Some xs" |
  "linear_term_impl xs (Fun f (t # ts)) = (case linear_term_impl xs t of 
    None \<Rightarrow> None
  | Some ys \<Rightarrow> linear_term_impl ys (Fun f ts))"

lemma [code]: "linear_term t = (case linear_term_impl [] t of None \<Rightarrow> False | Some _ \<Rightarrow> True)" 
proof -
  {
    fix xs ys
    have "(linear_term_impl xs t = None \<longrightarrow> (set xs \<inter> vars_term t \<noteq> {} \<or> \<not> linear_term t)) \<and> 
      (linear_term_impl xs t = Some ys \<longrightarrow> (set ys = set xs \<union> vars_term t) \<and> set xs \<inter> vars_term t = {} \<and> linear_term t)"
    proof (induct arbitrary: ys rule: linear_term_impl.induct, force, simp add: is_partition_Nil)
      case (3 xs f t ts zs)
      show ?case 
      proof (cases "linear_term_impl xs t")
        case None
        with 3 show ?thesis by auto
      next
        case (Some ys)
        note some = this
        with 3 have rec1: "set ys = set xs \<union> vars_term t \<and> set xs \<inter> vars_term t = {} \<and> linear_term t" by auto
        show ?thesis 
        proof (cases "linear_term_impl ys (Fun f ts)")
          case None 
          with rec1 Some have res: "linear_term_impl xs (Fun f (t # ts)) = None" by simp
          from None 3(2) Some have rec2: "set ys \<inter> vars_term (Fun f ts) \<noteq> {} \<or> \<not> linear_term (Fun f ts)" by simp
          hence "set xs \<inter> vars_term (Fun f (t # ts)) \<noteq> {} \<or> \<not> linear_term (Fun f (t # ts))" 
          proof 
            assume "set ys \<inter> vars_term (Fun f ts) \<noteq> {}" 
            then obtain x where x1: "x \<in> set ys" and x2: "x \<in> vars_term (Fun f ts)" by auto
            show ?thesis 
            proof (cases "x \<in> set xs")
              case True
              with x2 show ?thesis by auto
            next
              case False
              with x1 rec1 have "x \<in> vars_term t" by auto
              with x2 have "\<not> linear_term (Fun f (t # ts))" by (auto simp: is_partition_Cons)
              thus ?thesis ..
            qed
          next
            assume "\<not> linear_term (Fun f ts)" hence "\<not> linear_term (Fun f (t # ts))" by (auto simp: is_partition_Cons)
            thus ?thesis ..
          qed            
          with res show ?thesis by auto
        next
          case (Some us)
          with some have res: "linear_term_impl xs (Fun f (t # ts)) = Some us" by auto          
          {
            assume us: "us = zs"
            from Some[simplified us] 3(2) some 
            have rec2: "set zs = set ys \<union> vars_term (Fun f ts) \<and> set ys \<inter> vars_term (Fun f ts) = {} \<and> linear_term (Fun f ts)" by auto
            from rec1 rec2  
            have part1: "set zs = set xs \<union> vars_term (Fun f (t # ts)) \<and> set xs \<inter> vars_term (Fun f (t # ts)) = {}" (is ?part1) by auto
            from rec1 rec2 have "vars_term t \<inter> vars_term (Fun f ts) = {}" and "linear_term t" and "linear_term (Fun f ts)" by auto
            hence "linear_term (Fun f (t # ts))" (is ?part2) by (auto simp: is_partition_Cons) 
            with part1 have "?part1 \<and> ?part2" ..
          }
          with res show ?thesis by auto
        qed
      qed
    qed
  } note main = this
  from main[of "[]"] show ?thesis by (cases "linear_term_impl [] t", auto)
qed

end
