theory TreeAutomata 
imports Trs RPO Util2 
begin


(* syntax similar to AFP-Entry of TA *)
datatype ('q,'f)ta_rule = TA_rule 'f "'q list" 'q ("_ _ \<rightarrow> _")

record ('q,'f)ta = 
  ta_final :: "'q list"
  ta_rules :: "('q,'f) ta_rule list"

fun rsym where "rsym (f qs \<rightarrow> q) = (f, length qs)"
fun rqs where "rqs (f qs \<rightarrow> q) = qs"
fun rq where "rq (f qs \<rightarrow> q) = q"


fun ta_res :: "('q,'f)ta \<Rightarrow> ('f,'q)term \<Rightarrow> 'q set"
where "ta_res TA (Var q) = {q}"
  | "ta_res TA (Fun f ts) = {q | q qs. (f qs \<rightarrow> q) \<in> set (ta_rules TA) \<and> length qs = length ts \<and> (\<forall> i < length ts. qs ! i \<in> (map (ta_res TA) ts) ! i)}"

fun ta_res_list :: "('q,'f)ta \<Rightarrow> ('f,'q)term \<Rightarrow> 'q list"
where "ta_res_list TA (Var q) = [q]"
  | "ta_res_list TA (Fun f ts) = (
        let rec = map (ta_res_list TA) ts;
            n = length ts;
            rules = filter (\<lambda> rule. rsym rule = (f,n)) (ta_rules TA);
            arules = filter (\<lambda> rule. let qs = rqs rule in list_all (\<lambda> i. qs ! i \<in> set (rec ! i)) [0 ..< length qs]) rules
           in remdups (map rq arules))"

lemma ta_res_list: "set (ta_res_list TA t) = ta_res TA t"
proof (induct t)
  case (Var x) thus ?case by simp
next
  case (Fun f ts)
  show ?case
  proof (rule set_eqI)
    fix q
    show "(q \<in> set (ta_res_list TA (Fun f ts))) = (q \<in> ta_res TA (Fun f ts))"
      (is "(q \<in> ?L) = (q \<in> ?S)")
    proof
      assume "q \<in> ?L"
      then obtain rule where mem: "rule \<in> set (ta_rules TA)"
        and q: "rq rule = q"
        and sym: "rsym rule = (f, length ts)"
        and args: "\<forall> i \<in> {0 ..< length (rqs rule)}.
             rqs rule ! i \<in> set (map (ta_res_list TA) ts ! i)"
        by (auto simp: Let_def list_all_iff)
      from q sym obtain qs where rule: "rule = (f qs \<rightarrow> q)" and qs: "length qs = length ts" by (cases rule, auto)
      show "q \<in> ?S"
      proof (unfold ta_res.simps, intro CollectI, rule exI[of _ q], rule exI[of _ qs], intro conjI, simp, simp add: mem[unfolded rule], simp add: qs, intro allI impI)
        fix i
        assume i: "i < length ts"
        hence mem: "ts ! i \<in> set ts" by auto
        from args[unfolded rule, simplified] i qs
        have "qs ! i \<in> set (ta_res_list TA (ts ! i))" by auto
        from this[unfolded Fun[OF mem]] i
        show "qs ! i \<in> map (ta_res TA) ts ! i" by simp
      qed
    next
      assume "q \<in> ?S"
      then obtain qs where rule: "f qs \<rightarrow> q \<in> set (ta_rules TA)"
        and len: "length qs = length ts"
        and args: "\<forall> i < length ts. qs ! i \<in> map (ta_res TA) ts ! i"
        by auto
      have sym: "q = rq (f qs \<rightarrow> q)" by simp
      show "q \<in> ?L"
      proof (simp add: Let_def list_all_iff, rule, rule sym, rule, 
          intro conjI, rule rule, simp add: len, rule, simp add: len)
        fix i
        assume i: "i < length ts"
        hence mem: "ts ! i \<in> set ts" by simp
        show "qs ! i \<in> set (ta_res_list TA (ts ! i))"
          unfolding Fun[OF mem] using args i by auto
      qed
    qed
  qed
qed

declare ta_res.simps[code del]

definition ta_lang :: "('q,'f)ta \<Rightarrow> ('f,'q)term set"
where "ta_lang TA = {t. ground t \<and> set (ta_final TA) \<inter> ta_res TA t \<noteq> {}}"


lemma [code_inline]: "(t \<in> ta_lang TA) = (ground t \<and> list_inter  (ta_final TA) (ta_res_list TA t) \<noteq> [])" unfolding ta_lang_def
  by (simp, unfold list_inter[of "ta_final TA" "ta_res_list TA t", unfolded ta_res_list[of TA t], symmetric], insert list_inter, best)

(* corresponds to lemma 32(i) *)
lemma ta_res_subst: assumes "\<And> x. x \<in> vars_term t \<Longrightarrow> \<tau> x \<in> ta_res TA (get_subst \<sigma> x)"
  and "q \<in> ta_res TA (map_vars \<tau> t)"
  shows "q \<in> ta_res TA (t \<cdot> \<sigma>)"
using assms
proof (induct t arbitrary: q)
  case (Var x)
  thus ?case by auto
next
  case (Fun f ts)
  from Fun(3)
  obtain qs where rule: "f qs \<rightarrow> q \<in> set (ta_rules TA)"
    and len: "length qs = length ts"
    and rec: "\<And>i. i < length ts \<Longrightarrow> qs ! i \<in> ta_res TA (map_vars \<tau> (ts ! i))"
    by auto
  let ?ts = "map (\<lambda> t. t \<cdot> \<sigma>) ts"
  {
    fix i
    assume i: "i < length ts" 
    from i have mem: "ts ! i \<in> set ts" by auto
    hence "vars_term (ts ! i) \<subseteq> vars_term (Fun f ts)" by auto 
    with Fun(1)[OF mem Fun(2) rec[OF i]] have "qs ! i \<in> ta_res TA (ts ! i \<cdot> \<sigma>)"
      by auto
  } note main = this
  have id: "Fun f ts \<cdot> \<sigma> = Fun f ?ts" by simp
  show ?case
    by (unfold id ta_res.simps, intro CollectI, rule exI[of _ q], rule exI[of _ qs], intro conjI, simp, rule rule, simp add: len, simp add: main)
qed

lemma subst_merge: 
  assumes part: "is_partition (map vars_term ts)"
  shows "\<exists> \<sigma>. \<forall> i < length ts. \<forall> x \<in> vars_term (ts ! i). \<sigma> x = \<tau> i x"
proof -
  let ?\<tau> = "\<lambda> x. \<tau>  (THE i. i < length ts \<and> x \<in> vars_term (ts ! i)) x"
  show ?thesis  
  proof (rule exI[of _ ?\<tau>], intro allI impI ballI)
    fix i x
    assume i: "i < length ts" and x: "x \<in> vars_term (ts ! i)"
    have the: "(THE i. i < length ts \<and> x \<in> vars_term (ts ! i)) = i" 
    proof
      show "i < length ts \<and> x \<in> vars_term (ts ! i)" using i x by auto
    next
      fix j
      assume j: "j < length ts \<and> x \<in> vars_term (ts ! j)"
      from part[unfolded is_partition_alt is_partition_alt_def,
        THEN spec[of _ i], THEN spec[of _ j]] j i x show "j = i"
        by auto
    qed        
    thus "?\<tau> x = \<tau> i x" by simp
  qed
qed

(* corresponds to lemma 32(ii) *)
lemma ta_res_subst_linear: assumes "linear_term t"
  and "q \<in> ta_res TA (t \<cdot> \<sigma>)"
  shows "\<exists> \<tau>. q \<in> ta_res TA (map_vars \<tau> t) \<and> (\<forall> x \<in> vars_term t. \<tau> x \<in> ta_res TA (get_subst \<sigma> x))"
using assms
proof (induct t arbitrary: q)
  case (Var x)
  thus ?case by auto
next
  case (Fun f ts)
  from Fun(3)
  obtain qs where rule: "f qs \<rightarrow> q \<in> set (ta_rules TA)"
    and len: "length qs = length ts"
    and rec: "\<And>i. i < length ts \<Longrightarrow> qs ! i \<in> ta_res TA (ts ! i \<cdot> \<sigma>)"
    by auto
  {
    fix i
    assume i: "i < length ts" 
    from i have mem: "ts ! i \<in> set ts" by auto
    from Fun(2) mem have "linear_term (ts ! i)" by (auto simp: list_all_iff)
    note Fun(1)[OF mem this rec[OF i]]
  }
  hence "\<forall> i. \<exists> \<tau>. i < length ts \<longrightarrow> qs ! i \<in> ta_res TA (map_vars \<tau> (ts ! i)) \<and>
             (\<forall> x \<in> vars_term (ts ! i). \<tau> x \<in> ta_res TA (get_subst \<sigma> x))"
    by blast
  from choice[OF this] obtain \<tau> where
    ind: "\<And> i. i < length ts \<Longrightarrow> qs ! i \<in> ta_res TA (map_vars (\<tau> i) (ts ! i)) \<and> 
             (\<forall> x \<in> vars_term (ts ! i). (\<tau> i) x \<in> ta_res TA (get_subst \<sigma> x))"
    by auto
  from Fun(2) have "is_partition (map vars_term ts)" by auto
  from subst_merge[OF this, of \<tau>] obtain \<tau>2 where \<tau>2: "\<And> i. i < length ts \<Longrightarrow> \<forall> x \<in> vars_term (ts ! i). \<tau>2 x = \<tau> i x" by auto
  show ?case
  proof (rule exI[of _ \<tau>2], intro conjI)
    show "\<forall> x \<in> vars_term (Fun f ts). \<tau>2 x \<in> ta_res TA (get_subst \<sigma> x)"
    proof
      fix x
      assume "x \<in> vars_term (Fun f ts)"
      then obtain t where "t \<in> set ts" and "x \<in> vars_term t" by auto
      then obtain i where i: "i < length ts" and x: "x \<in> vars_term (ts ! i)" unfolding set_conv_nth by auto
      from ind[OF i] \<tau>2[OF i] x show "\<tau>2 x \<in> ta_res TA (get_subst \<sigma> x)" by auto
    qed
  next
    { 
      fix i
      assume i: "i < length ts"
      have id: "map_vars \<tau>2 (ts ! i) = map_vars (\<tau> i) (ts ! i)" 
        by (rule map_vars_vars_term, insert \<tau>2[OF i], auto)
      from ind[OF i, THEN conjunct1] 
      have "qs ! i \<in> ta_res TA (map_vars \<tau>2 (ts ! i))" unfolding id .
    } note main = this
    have "q \<in> ta_res TA (Fun f (map (map_vars \<tau>2) ts))" 
      by (unfold ta_res.simps, intro CollectI, rule exI[of _ q], rule exI[of _ qs], intro conjI, simp, rule rule, simp add: len, simp add: main)
    thus "q \<in> ta_res TA (map_vars \<tau>2 (Fun f ts))" by auto
  qed
qed

definition ta_rqs :: "('q,'f)ta \<Rightarrow> 'q set"
where "ta_rqs TA \<equiv> rq ` set (ta_rules TA)"

definition ta_rqs_list :: "('q,'f)ta \<Rightarrow> 'q list"
where "ta_rqs_list TA \<equiv> remdups (map rq (ta_rules TA))"

lemma ta_rqs_list: "set (ta_rqs_list TA) = ta_rqs TA"
  unfolding ta_rqs_def ta_rqs_list_def by auto

lemma ta_rqs_res: assumes ground: "ground t" 
  shows "ta_res TA t \<subseteq> ta_rqs TA"
proof
  fix q 
  assume q: "q \<in> ta_res TA t"
  from ground obtain f ts where t: "t = Fun f ts" by (cases t, auto)
  from q[unfolded t] obtain qs where "(f qs \<rightarrow> q) \<in> set (ta_rules TA)" by auto
  thus "q \<in> ta_rqs TA" unfolding ta_rqs_def by force
qed    


(* the restrictions on vars_term and \<tau> are made to ensure that we will only
   have to test finitely many \<tau> *)
definition ta_trs_compatible :: "('q,'f)ta \<Rightarrow> ('f,'v)trs \<Rightarrow> bool"
where "ta_trs_compatible TA R \<equiv> \<forall> l r \<tau>. (l,r) \<in> R \<and> vars_term r \<subseteq> vars_term l  \<and> (\<tau> ` vars_term l \<subseteq> ta_rqs TA) \<longrightarrow> 
  ta_res TA (map_vars \<tau> l) \<subseteq> ta_res TA (map_vars \<tau> r)"

(* TODO: the following test does brute force enumeration over
  all possible substitutions; moreover fetching an individual assignment
  is linear; this might be a bottle-neck and both situtations can be improved *)
definition ta_trs_compatible_list :: "('q,'f)ta \<Rightarrow> ('f,'v)rules \<Rightarrow> bool"
where "ta_trs_compatible_list TA R \<equiv> let
     Q = ta_rqs_list TA in
     list_all (\<lambda> (l,r). list_all (\<lambda> vec. let \<tau> = assign_from_vector vec 
                            in  set (ta_res_list TA (map_vars \<tau> l)) \<subseteq>
                                set (ta_res_list TA (map_vars \<tau> r)))
                        (enum_vectors Q (vars_term_list l))
     ) R"

(* I believe that also equivalence holds but was to lazy to prove it *)
lemma ta_trs_compatible_list: assumes wf_trs: "\<And> l r. (l,r) \<in> set R \<Longrightarrow> vars_term r \<subseteq> vars_term l"
  and non_empty: "ta_rules TA \<noteq> []"
  and comp: "ta_trs_compatible_list TA R"
  shows "ta_trs_compatible TA (set R)" 
unfolding ta_trs_compatible_def 
proof(intro allI impI, clarify)
  fix l r \<tau> q
  assume q: "q \<in> ta_res TA (map_vars \<tau> l)"
  and rule: "(l,r) \<in> set R"
  and wf: "vars_term r \<subseteq> vars_term l"
  and \<tau>: "\<tau> ` vars_term l \<subseteq> ta_rqs TA"
  let ?Q = "ta_rqs_list TA"
  from non_empty have "?Q \<noteq> []" unfolding ta_rqs_list_def by (cases "ta_rules TA", auto)
  from enum_vectors_complete[OF this, of "vars_term_list l"]
  obtain vec where vec: "vec \<in> set (enum_vectors (ta_rqs_list TA) (vars_term_list l))" and \<tau>2: "\<And> x c. x \<in> set (vars_term_list l) \<Longrightarrow> c \<in> set (ta_rqs_list TA) \<Longrightarrow>  \<tau> x = c \<longrightarrow> assign_from_vector vec x = c" by blast
  let ?\<tau> = "assign_from_vector vec"
  {
    fix x
    assume x: "x \<in> vars_term l"
    with \<tau> have "\<tau> x \<in> ta_rqs TA" by auto    
    from \<tau>2[unfolded vars_term_list_sound ta_rqs_list, OF x this] have "\<tau> x = ?\<tau> x"
      by simp
  } note \<tau>2 = this
  have l: "map_vars \<tau> l = map_vars ?\<tau> l"
    by (rule map_vars_vars_term[OF \<tau>2])
  have r: "map_vars \<tau> r = map_vars ?\<tau> r"
    by (rule map_vars_vars_term, insert \<tau>2 wf, auto)
  from comp[unfolded ta_trs_compatible_list_def Let_def list_all_iff ta_res_list,
     THEN bspec[OF _ rule]] vec q
  show "q \<in> ta_res TA (map_vars \<tau> r)" unfolding l r by auto
qed


lemma ta_trs_compatible_res: assumes comp: "ta_trs_compatible TA R"
  and ll: "left_linear_trs R"
  and wf_trs: "\<And> l r. (l,r) \<in> R \<Longrightarrow> vars_term r \<subseteq> vars_term l"
  and step: "(s,t) \<in> rstep R"
  and ground: "ground s"
  shows "ta_res TA s \<subseteq> ta_res TA t"
  using step ground
proof (induct)
  case (IH C \<sigma> l r)
  hence ground: "ground (l \<cdot> \<sigma>)" by simp
  show ?case
  proof (induct C)
    case (More f bef C aft)
    let ?args = "\<lambda> l. bef @ C\<langle>l \<cdot> \<sigma>\<rangle> # aft"
    let ?l = "Fun f (?args l)"
    let ?r = "Fun f (?args r)"
    let ?n = "Suc (length bef + length aft)"
    let ?m = "length bef"
    show ?case
    proof
      fix q
      assume "q \<in> ta_res TA (More f bef C aft)\<langle>l\<cdot>\<sigma>\<rangle>"
      hence "q \<in> ta_res TA ?l" by simp
      from this[unfolded ta_res.simps] obtain qs where 
        rule: "f qs \<rightarrow> q \<in> set (ta_rules TA)"
        and len: "length qs = ?n"
        and rec: "\<forall>i < ?n. qs ! i \<in> map (ta_res TA) (?args l) ! i"
        by auto
      have len2: "\<And>t. ?m < length (?args t)" by auto
      from rec[THEN spec, of ?m, unfolded nth_map[OF len2]]
      have "qs ! ?m \<in> ta_res TA (C\<langle>l \<cdot> \<sigma>\<rangle>)"  by auto
      with More have m: "qs ! ?m \<in> ta_res TA (C\<langle>r\<cdot>\<sigma>\<rangle>)" by auto
      { 
        fix i
        assume i: "i < ?n"
        have "qs ! i \<in> map (ta_res TA) (?args r) ! i" 
        proof (cases "i = ?m")
          case True
          show ?thesis using m unfolding True
            by (simp add: nth_append)
        next
          case False
          note rec = rec[THEN spec, THEN mp, OF i]
          show ?thesis 
          proof (cases "i < ?m")
            case True
            show ?thesis using rec
              by (simp add: nth_append True)
          next
            case False
            with `i \<noteq> ?m` have "i - ?m > 0" by auto
            then obtain j where im: "i - ?m = Suc j" by (cases "i - ?m", auto)
            show ?thesis using rec 
              by (simp add: nth_append False im)
          qed
        qed
      }
      hence rec: "\<forall>i < ?n. qs ! i \<in> map (ta_res TA) (?args r) ! i"
        by simp        
      have "q \<in> ta_res TA ?r" 
        unfolding ta_res.simps
        by (intro CollectI, rule exI[of _ q], rule exI[of _ qs],
          insert rule len rec, simp)
      thus "q \<in> ta_res TA (More f bef C aft)\<langle>r\<cdot>\<sigma>\<rangle>" by simp
    qed
  next
    case Hole
    show ?case
    proof (unfold ctxt_apply.simps, rule)
      fix q
      assume q: "q \<in> ta_res TA (l \<cdot> \<sigma>)"
      from ll IH(1) have "linear_term l" unfolding left_linear_trs_def by auto
      from ta_res_subst_linear[OF this q] obtain \<tau> where
        q: "q \<in> ta_res TA (map_vars \<tau> l)" and 
        cond: "\<forall> x \<in> vars_term l. \<tau> x \<in> ta_res TA (get_subst \<sigma> x)" by auto
      have "\<tau> ` vars_term l \<subseteq> ta_rqs TA" 
      proof 
        fix q
        assume "q \<in> \<tau> ` vars_term l"
        then obtain x where x: "x \<in> vars_term l" and q: "q = \<tau> x" by auto
        with cond have q: "q \<in> ta_res TA (get_subst \<sigma> x)" by auto
        from ground x have "ground (get_subst \<sigma> x)" by auto
        from ta_rqs_res[OF this] q show "q \<in> ta_rqs TA" by auto
      qed
      with IH(1) wf_trs[OF IH(1)] have "(l,r) \<in> R \<and> vars_term r \<subseteq> vars_term l \<and> (\<tau> ` vars_term l \<subseteq> ta_rqs TA)" by auto
      from comp[unfolded ta_trs_compatible_def, 
        THEN spec, THEN spec, THEN spec, THEN mp[OF _ this]]
        q have q: "q \<in> ta_res TA (map_vars \<tau> r)" by auto
      from ta_res_subst[OF _ q] wf_trs[OF IH(1)] cond
      show "q \<in> ta_res TA (r \<cdot> \<sigma>)" by force
    qed
  qed
qed


lemma ta_trs_compatible_lang: assumes comp: "ta_trs_compatible TA R"
  and ll: "left_linear_trs R"
  and wf_trs: "\<And> l r. (l,r) \<in> R \<Longrightarrow> vars_term r \<subseteq> vars_term l"
  and step: "(s,t) \<in> rstep R"
  and s: "s \<in> ta_lang TA"
  shows "t \<in> ta_lang TA"
proof -
  from s[unfolded ta_lang_def]
  have g: "ground s" and acc: "set (ta_final TA) \<inter> ta_res TA s \<noteq> {}" by auto
  from rstep_ground[OF wf_trs g step] have gt: "ground t" .
  from ta_trs_compatible_res[OF comp ll wf_trs step g] acc 
  have acc: "set (ta_final TA) \<inter> ta_res TA t \<noteq> {}" by auto
  from gt acc show "t \<in> ta_lang TA" unfolding ta_lang_def by auto
qed

definition ta_contains :: "('f \<times> nat) list \<Rightarrow> ('q,'f)ta \<Rightarrow> 'q \<Rightarrow> bool"
  where "ta_contains F TA q \<equiv> let rules = ta_rules TA in
  q \<in> set (ta_final TA) \<and> list_all (\<lambda> (f,n). (f (replicate n q) \<rightarrow> q) \<in> set rules) F"

lemma ta_contains: fixes TA :: "('q,'f)ta"
  assumes contain: "ta_contains F TA q"
  shows "(wf_terms (set F) :: ('f,'q)term set) \<inter> {t. ground t} \<subseteq> ta_lang TA"
proof (clarify)
  note contain = contain[unfolded ta_contains_def Let_def list_all_iff]
  fix t :: "('f,'q)term"
  assume "t \<in> wf_terms (set F)" and g: "ground t"
  hence "q \<in> ta_res TA t"
  proof (induct)
    case (Var x) thus ?case by simp
  next
    case (Fun f ts)
    let ?qs = "replicate (length ts) q"
    from contain Fun(1) have rule: "f ?qs \<rightarrow> q \<in> set (ta_rules TA)" by auto
    show ?case unfolding ta_res.simps
    proof (rule, rule exI[of _ q], rule exI[of _ ?qs], intro conjI, simp,
      rule rule, simp, intro allI impI, simp)
      fix i
      assume "i < length ts"
      hence mem: "ts ! i \<in> set ts" by simp
      from Fun(3)[OF mem] Fun(4) mem
      show "q \<in> ta_res TA (ts ! i)" by auto
    qed
  qed
  with g contain show "t \<in> ta_lang TA" unfolding ta_lang_def by auto
qed  


(* lemma 33 *)
lemma ta_trs_compatible: assumes comp: "ta_trs_compatible TA R"
  and ll: "left_linear_trs R"
  and wf_trs: "\<And> l r. (l,r) \<in> R \<Longrightarrow> vars_term r \<subseteq> vars_term l"
  and L: "L \<subseteq> ta_lang TA"
  shows "{t | s t. s \<in> L \<and> (s,t) \<in> (rstep R)^*} \<subseteq> ta_lang TA"
proof -
  {
    fix s t
    assume "(s,t) \<in> (rstep R)^*" and "s \<in> L" 
    hence "t \<in> ta_lang TA"
    proof (induct)
      case base
      thus ?case using L by auto
    next
      case (step t u)
      hence t: "t \<in> ta_lang TA" by simp
      show ?case
        by (rule ta_trs_compatible_lang[OF comp ll wf_trs step(2) t])
    qed
  }
  thus ?thesis by auto
qed

definition compact :: "('f,'v)term set \<Rightarrow> ('f,'v)trs \<Rightarrow> bool"
  where "compact L R \<equiv> \<exists> F. finite F \<and> {t | s t. s \<in> L \<and> (s,t) \<in> (rstep R)^*} \<subseteq> wf_terms F"


definition ta_sig :: "('q,'f)ta \<Rightarrow> 'f sig"
where "ta_sig TA \<equiv> rsym ` (set (ta_rules TA))"

lemma ta_sig_res: "q \<in> ta_res TA t \<Longrightarrow> t \<in> wf_terms (ta_sig TA)"
proof (induct t arbitrary: q)
  case (Var x)
  show ?case ..
next
  case (Fun f ts)
  from Fun(2)
  obtain qs where rule: "f qs \<rightarrow> q \<in> set (ta_rules TA)"
    and len: "length qs = length ts"
    and rec: "\<forall> i < length ts. qs ! i \<in> ta_res TA (ts ! i)"
    by auto
  {
    fix t 
    assume mem: "t \<in> set ts"
    then obtain i where i: "i < length ts"
      and t: "t = ts ! i" unfolding set_conv_nth by auto
    from Fun(1)[OF mem[unfolded t] rec[THEN spec, THEN mp[OF _ i]]]
    have "t \<in> wf_terms (ta_sig TA)" unfolding t by auto
  } note rec = this
  from rule len have f: "(f,length ts) \<in> ta_sig TA" unfolding ta_sig_def
    by force
  show ?case
    by (rule, rule f, rule rec)
qed

lemma ta_sig_lang: "ta_lang TA \<subseteq> wf_terms (ta_sig TA)"
proof
  fix t
  assume "t \<in> ta_lang TA"
  then obtain q where "q \<in> ta_res TA t" unfolding ta_lang_def
    by auto
  from ta_sig_res[OF this] show "t \<in> wf_terms (ta_sig TA)" .
qed

lemma finite_ta_sig: "finite (ta_sig TA)"
  unfolding ta_sig_def by auto
    

(* lemma 35 *)
lemma ta_compact: assumes comp: "ta_trs_compatible TA R"
  and ll: "left_linear_trs R"
  and wf_trs: "\<And> l r. (l,r) \<in> R \<Longrightarrow> vars_term r \<subseteq> vars_term l"
  and L: "L \<subseteq> ta_lang TA"
  shows "compact L R"
  unfolding compact_def
  by (rule exI[of _ "ta_sig TA"], intro conjI[OF finite_ta_sig], rule subset_trans[OF ta_trs_compatible[OF comp ll wf_trs L] ta_sig_lang])

definition locally_terminating :: "('f,'v)trs \<Rightarrow> bool"
  where "locally_terminating R \<equiv> \<forall> F. finite F \<longrightarrow> SN {(s,t). (s,t) \<in> rstep R \<and> s \<in> wf_terms F \<and> t \<in> wf_terms F}"

(* lemma 20 *)
lemma locally_terminating_compact: 
  assumes lterm: "locally_terminating R"
  and compact: "compact L R"
  and t: "t \<in> L"
  shows "SN_elt (rstep R) t"
proof -
  from compact[unfolded compact_def] t obtain F where
    F: "finite F"
    and wf_terms: "{s. (t,s) \<in> (rstep R)^*} \<subseteq> wf_terms F" (is "?L \<subseteq> _")
    by auto
  from lterm[unfolded locally_terminating_def] F
  have SN: "SN {(s,t). (s,t) \<in> rstep R \<and> s \<in> wf_terms F \<and> t \<in> wf_terms F}" (is "SN ?R")
    by auto
  show "SN_elt (rstep R) t"
  proof
    fix f 
    assume t: "f 0 = t" and steps: "\<forall>i. (f i, f (Suc i)) \<in> rstep R" 
    {
      fix i
      have "f i \<in> ?L"
      proof (induct i)
        case 0
        thus ?case unfolding t by auto
      next
        case (Suc n)
        thus ?case using steps[THEN spec[of _ n]] by auto
      qed
      with wf_terms have "f i \<in> wf_terms F" by auto
    } note wf_terms = this
    {
      fix i
      have "(f i, f (Suc i)) \<in> ?R" 
        using wf_terms[of i] wf_terms[of "Suc i"] steps by auto
    }
    with SN show False unfolding SN_defs by auto
  qed    
qed

text {* here the part about matchbounds starts *}



locale enrichment = 
  fixes R :: "('f,'v)trs"
    and R' :: "('g,'v)trs"
    and lift :: "('f,'v)term \<Rightarrow> ('g,'v)term"
    and base :: "('g,'v)term \<Rightarrow> ('f,'v)term"
 assumes base_lift: "\<And> s :: ('f,'v)term. base (lift s) = s"
   and lift: "\<And> s t s'. (s,t) \<in> rstep R \<Longrightarrow> s = base s' \<Longrightarrow> \<exists> t'. t = base t' \<and> (s',t') \<in> rstep R'"
begin


lemma enrich2: 
  "{t | t s. s \<in> L \<and> (s,t) \<in> (rstep R)^*} \<subseteq> base ` {t'. \<exists> s'. s' \<in> lift ` L \<and> (s',t') \<in> (rstep R')^* }" (is "_ \<subseteq> ?R'")
proof -
  {
    fix t s
    assume s: "s \<in> L" and steps: "(s,t) \<in> (rstep R)^*"
    from steps s 
    have "t \<in> ?R'"
    proof (induct)
      case base
      from base_lift[of s] base
      show ?case by force
    next
      case (step t u)
      from step(3)[OF step(4)] have "t \<in> ?R'" .
      then obtain t' s' where s': "s' \<in> lift ` L"
        and t: "t = base t'"
        and steps: "(s',t') \<in> (rstep R')^*" by auto
      from lift[OF step(2) t] obtain u' where u: "u = base u'"
        and step: "(t',u') \<in> rstep R'" by auto
      show ?case
        by (rule, rule u, rule, rule, intro conjI, rule s', insert steps step, auto)
    qed
  }
  thus ?thesis by auto
qed

fun sequence_gen :: "('g,'v)term \<Rightarrow> (nat \<Rightarrow> ('f,'v)term) \<Rightarrow> nat \<Rightarrow> ('g,'v)term" 
  where "sequence_gen s' f 0 = s'"
  | "sequence_gen s' f (Suc i) = (SOME t'. f (Suc i) = base t' \<and> (sequence_gen s' f i, t') \<in> rstep R' )"

lemma enrichment3: assumes LL: "lift ` L \<subseteq> L'"
  and SN: "\<And> t'. t' \<in> L' \<Longrightarrow> SN_elt (rstep R') t'"
  and tL: "t \<in> L"
  shows "SN_elt (rstep R) t"
proof
  fix f
  assume t: "f 0 = t" and steps: "\<forall> i. (f i, f (Suc i)) \<in> rstep R"
  hence steps: "\<And> i. (f i, f (Suc i)) \<in> rstep R" by auto
  from LL tL have tL': "lift t \<in> L'" by auto
  from base_lift[of t]  tL have bl: "base (lift t) = t" by auto
  {
    fix i
    have "f (Suc i) = base (sequence_gen (lift t) f (Suc i)) \<and> (sequence_gen (lift t) f i, sequence_gen (lift t) f (Suc i)) \<in> rstep R'"
    proof (induct i)
      case 0
      thus ?case 
        by (simp add: bl t, insert someI_ex[OF lift[OF steps[of 0, unfolded t] bl[symmetric]]], auto)
    next
      case (Suc i)
      from someI_ex[OF lift[OF steps[of "Suc i"] Suc[THEN conjunct1]]]
      show ?case by simp
    qed      
    note this[THEN conjunct2]
  } note steps = this
  have "sequence_gen (lift t) f 0 \<in> L'" by (simp add: tL')
  with steps SN[OF this] show False unfolding SN_elt_def
    by blast
qed
end


fun base :: "'f \<times> nat \<Rightarrow> 'f" where "base (f,h) = f"
fun height :: "'f \<times> nat \<Rightarrow> nat" where "height (f,h) = h"
fun lift :: "nat \<Rightarrow> 'f \<Rightarrow> 'f \<times> nat" where "lift h f = (f,h)"
abbreviation base_term where "base_term \<equiv> map_term base"
abbreviation lift_term where "lift_term h \<equiv> map_term (lift h)"

lemma base_lift: "base_term (lift_term h s) = s"
  by (simp only: map_term_comp o_def, simp add: map_term_id[unfolded id_def])

declare size_simps[simp]
fun map_match :: "('f,'v)term \<Rightarrow> ('g,'v)term \<Rightarrow> ('g,'v)term"
where "map_match (Var x) s' = (Var x)"
  |   "map_match (Fun f ls) (Fun g ss') = (Fun g (map (\<lambda> (l,s'). map_match l s') (zip ls ss')))"
declare size_simps[simp del]


lemma vars_term_map_match: "map_term fg s' = s \<cdot> \<sigma> \<Longrightarrow> vars_term (map_match s s') = vars_term s"
proof (induct s arbitrary: s', simp)
  case (Fun f ss)
  then obtain g ss' where s': "s' = Fun g ss'" by (cases s', auto)
  from Fun(2)[unfolded s'] have ss': "map (map_term fg) ss' = map (\<lambda> t. t \<cdot> \<sigma>) ss" by (auto)
  from ss' have "length (map (map_term fg) ss') = length (map (\<lambda> t. t \<cdot> \<sigma>) ss)" by simp
  hence len: "length ss = length ss'" by simp
  {
    fix i
    assume i: "i < length ss"
    hence mem: "ss ! i \<in> set ss" by auto
    from map_nth_conv[OF ss'] i len have "map_term fg (ss' ! i) = ss ! i \<cdot> \<sigma>" by auto
    with Fun(1)[unfolded s', OF mem this] have "vars_term (ss ! i) = vars_term (map_match (ss ! i) (ss' ! i))" by auto
  } note main = this
  have id: "set ss = set (map (\<lambda> i. ss ! i) [0 ..< length ss])" 
    using map_nth[of ss] by auto
  show ?case unfolding s' 
    by (simp add: zip_nth_conv[OF len] id, insert main len, auto)
qed



(* lemma 4, existence *)
lemma map_term_linear_ex: assumes lin: "linear_term l"
  and inst: "map_term fg s' = l \<cdot> \<sigma>"
  shows "map_term fg (map_match l s') = l \<and> (\<exists> \<tau>. s' = (map_match l s') \<cdot> \<tau>)"
using assms
proof (induct l arbitrary: s')
  case (Var x)
  show ?case
    by (simp, rule exI[of _ "Subst (\<lambda> _. s')"], simp)
next
  case (Fun f ss)
  from Fun(3) obtain f' ss' where s': "s' = Fun f' ss'" and 
    f': "fg f' = f" and ss': "map (map_term fg) ss' = map (\<lambda> t. t \<cdot> \<sigma>) ss" by (cases s', auto)
  from ss' have "length (map (map_term fg) ss') = length (map (\<lambda> t. t \<cdot> \<sigma>) ss)" by simp
  hence len: "length ss = length ss'" by simp
  have l2: "length [0 ..< length ss] = length ss" by auto
  have l3: "length [0 ..< length ss'] = length ss'" by auto
  {
    fix i
    assume i: "i < length ss"
    hence mem: "ss ! i \<in> set ss" by auto
    with Fun(2) have lin: "linear_term (ss ! i)" by (auto simp: list_all_iff)
    from map_nth_conv[OF ss'] i len have "map_term fg (ss' ! i) = ss ! i \<cdot> \<sigma>" by auto
    note Fun(1)[OF mem lin this] this
  } note ind = this
  let ?P = "\<lambda> i \<tau>. map_term fg (map_match (ss ! i) (ss' ! i)) = ss ! i \<and> 
                  ss' ! i = map_match (ss ! i) (ss' ! i) \<cdot> \<tau>"
  from ind have "\<forall> i. \<exists> \<tau>. i < length ss \<longrightarrow> ?P i \<tau>" by blast
  from choice[OF this] obtain \<tau> where \<tau>: "\<And> i. i < length ss \<Longrightarrow> ?P i (\<tau> i)" by auto
  from Fun(2) have part_ss: "is_partition (map vars_term ss)" by  auto
  have "map vars_term ss = map vars_term (map (\<lambda> i. map_match (ss ! i) (ss' ! i)) [0 ..< length ss])"
  proof (simp add: o_def, rule nth_map_conv, simp, simp, intro allI impI)
    fix i
    assume i: "i < length ss"
    from vars_term_map_match[OF ind(2)[OF i]]
    show "vars_term (ss ! i) = vars_term (map_match (ss ! i) (ss' ! i))" ..
  qed
  from subst_merge[OF part_ss[unfolded this], of "\<lambda> i. get_subst (\<tau> i)"]
  obtain \<tau>2 where \<tau>2: "\<And> i x. i < length ss \<Longrightarrow> x \<in> vars_term (map_match (ss ! i) (ss' ! i)) \<Longrightarrow> \<tau>2 x = get_subst (\<tau> i) x" by auto
  show ?case unfolding s' 
  proof (intro conjI, simp add: f' o_def zip_nth_conv[OF len] len[symmetric], unfold map_nth_eq_conv[OF l2], simp add: ind)
    show "\<exists> \<tau>. Fun f' ss' = map_match (Fun f ss) (Fun f' ss') \<cdot> \<tau>"
    proof (simp add: o_def zip_nth_conv[OF len], unfold eq_commute[of ss']
      map_nth_eq_conv[OF l3], simp)
      show "\<exists> \<tau>. \<forall>i < length ss'. map_match (ss ! i) (ss' ! i) \<cdot> \<tau> = ss' ! i"
      proof (rule exI[of _ "Subst \<tau>2"], intro allI impI)
        fix i
        assume "i < length ss'"
        with len have i: "i < length ss" by auto
        have "map_match (ss ! i) (ss' ! i) \<cdot> Subst \<tau>2 = 
              map_match (ss ! i) (ss' ! i) \<cdot> \<tau> i" 
          by (rule term_subst_eq, insert \<tau>2[OF i], auto)
        also have "... = ss' ! i" by (rule \<tau>[OF i, THEN conjunct2, symmetric])
        finally show "map_match (ss ! i) (ss' ! i) \<cdot> Subst \<tau>2  = ss' ! i" .
      qed
    qed
  qed
qed

(* lemma 4, uniqueness *)
lemma map_term_linear_unique: assumes lin: "linear_term l"
  and inst: "map_term fg s' = l \<cdot> \<sigma>"
  and base: "map_term fg t' = l"
  and match: "s' = t' \<cdot> \<tau>"
  shows "t' = map_match l s'"
using assms 
proof (induct l arbitrary: s' t')
  case (Var x)
  hence "t' = Var x" by (cases t', auto)
  with Var show ?case by auto
next
  case (Fun f ss)
  from Fun(3)
  obtain f' ss' where s': "s' = Fun f' ss'" and 
    f': "fg f' = f" and ss': "map (map_term fg) ss' = map (\<lambda> t. t \<cdot> \<sigma>) ss" by (cases s', auto)
  from ss' have "length (map (map_term fg) ss') = length (map (\<lambda> t. t \<cdot> \<sigma>) ss)" by simp
  hence len: "length ss = length ss'" by simp
  from Fun(4) obtain g' ts' where t': "t' = Fun g' ts'" by (cases t', auto)
  from Fun(5)[unfolded s' t'] t' have t': "t' = Fun f' ts'" and len2: "length ss' = length ts'" and ts': "ss' = map (\<lambda> t. t \<cdot> \<tau>) ts'" by auto
  have len3: "length [0 ..< length ts' ] = length ts'" by auto
  {
    fix i
    assume  "i < length ts'"
    hence i: "i < length ss" using len len2 by simp
    hence mem: "ss ! i \<in> set ss" by auto
    with Fun(2) have lin: "linear_term (ss ! i)" by (auto simp: list_all_iff)
    from i len map_nth_conv[OF ss'] 
    have base: "map_term fg (ss' ! i) = ss ! i \<cdot> \<sigma>" by auto
    from ts'[unfolded eq_commute[of ss'] map_nth_eq_conv[OF len2[symmetric]]]
      i len have ssi: "ss' ! i = ts' ! i \<cdot> \<tau>" by auto
    from Fun(4)[unfolded t'] have id: "map (map_term fg) ts' = ss" by simp
    from len len2 have "length ts' = length ss" by simp
    from id[unfolded map_nth_eq_conv[OF this]] i have base2: "map_term fg (ts' ! i) = ss ! i" by simp
    have "ts' ! i = map_match (ss ! i) (ss' ! i)" 
      by (rule Fun(1)[OF mem lin base base2 ssi])
  } note ind = this 
  show ?case unfolding t' s' 
    by (simp add: zip_nth_conv[OF len] o_def eq_commute[of ts'] len2
      map_nth_eq_conv[OF len3], insert ind, auto)
qed


fun sym_collect :: "(('f,'v)term \<Rightarrow> bool) \<Rightarrow> ('f,'v)term \<Rightarrow> 'f list"
where "sym_collect p (Var x) = []"
  |   "sym_collect p (Fun f ts) = (if p (Fun f ts) then [f] else []) @ 
  concat (map (sym_collect p) ts)"

definition cover :: "(('f,'v)rule \<Rightarrow> ('f,'v)term \<Rightarrow> bool) \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f \<times> nat,'v)trs"
  where "cover ff R \<equiv> { (l', lift_term h r) | l r l' h. 
       (l,r) \<in> R 
     \<and> base_term l' = l 
     \<and> h = Suc (min_list (map height (sym_collect (\<lambda> t'. ff (l,r) (base_term t')) l')))}"

lemma cover6a: "(l',r') \<in> cover ff {(l,r)} \<Longrightarrow> (base_term l',base_term r') = (l,r)"
  unfolding cover_def
  by (auto simp: base_lift)
  
lemma cover6b: assumes step: "(s',t') \<in> rstep (cover ff R)"
  shows "(base_term s',base_term t') \<in> rstep R"
  using assms
proof (induct)
  case (IH C \<sigma> l' r')
  then obtain l r where lr: "(l,r) \<in> R" and l'r': "(l',r') \<in> cover ff {(l,r)}"
    unfolding cover_def by auto
  from lr cover6a[OF l'r'] have lr: "(base_term l',base_term r') \<in> R" by auto
  show ?case
    by (simp, rule, rule lr, auto)
qed

(* proposition 7 extended (the original formulation does not suffice
   to get lemma 13) *)
lemma enrichment_cover_extended: assumes lin: "linear_term l"
   and wf: "vars_term r \<subseteq> vars_term l"
   and s': "base_term s' = C\<langle>l \<cdot> \<sigma>\<rangle>"
   shows "\<exists> l' r' \<tau> D. (l',r') \<in> cover ff {(l,r)} \<and> s' = D\<langle>l' \<cdot> \<tau>\<rangle> \<and> C = map_ctxt base D \<and> l = base_term l' \<and> r = base_term r' \<and> (\<forall> x \<in> vars_term r. get_subst (map_subst base \<tau>) x = get_subst \<sigma> x)"
proof -
  from map_term_ctxt_decomp[OF s'] obtain D u' where
      D: "C = map_ctxt base D" and base: "base_term u' = l \<cdot> \<sigma>" 
      and s': "s' = D\<langle>u'\<rangle>" by auto
  from map_term_linear_ex[OF lin base] obtain l' \<tau> where l: "l = base_term l'"
      and u': "u' = l' \<cdot> \<tau>" by force
  from l obtain r' where coverlr: "(l',r') \<in> cover ff {(l,r)}" 
    unfolding cover_def by auto
  from cover6a[OF coverlr] have r: "r = base_term r'" by auto      
  show ?thesis unfolding s' u' 
  proof (rule exI[of _ l'], rule exI[of _ r'], rule exI[of _ \<tau>], rule exI[of _ D], simp add: coverlr D, simp add: l r, intro ballI)
    fix x
    assume "x \<in> vars_term r'"
    thus "get_subst (map_subst base \<tau>) x = get_subst \<sigma> x" 
      using term_subst_eq_rev[OF base[unfolded l s' u', simplified]] wf
      unfolding l r vars_map_term2 by auto
  qed
qed

lemma enrichment_cover: assumes ll: "left_linear_trs R"
  and wf: "\<And> l r. (l,r) \<in> R \<Longrightarrow> vars_term r \<subseteq> vars_term l"
  shows "enrichment R (cover ff R) (lift_term n) base_term"
proof
  fix t
  show "base_term (lift_term n t) = t" by (rule base_lift)
next
  fix s t s'
  assume step: "(s,t) \<in> rstep R" and s: "s = base_term s'"
  thus "\<exists> t'. t = base_term t' \<and> (s',t') \<in> rstep (cover ff R)"
  proof (induct arbitrary: s')
    case (IH C \<sigma> l r s')
    from IH(1) ll have lin:"linear_term l" unfolding left_linear_trs_def by auto
    from IH(2) have s': "base_term s' = C\<langle>l \<cdot> \<sigma>\<rangle>" by simp    
    from enrichment_cover_extended[OF lin wf[OF IH(1)] s']
    obtain l' r' \<tau> D where rule: "(l',r') \<in> cover ff {(l,r)}"
      and s': "s' = D\<langle>l' \<cdot> \<tau>\<rangle>" and D: "C = map_ctxt base D"
      and l: "l = base_term l'" and r: "r = base_term r'"
      and \<tau>: "\<And> x. x \<in> vars_term r \<Longrightarrow> get_subst (map_subst base \<tau>) x = get_subst \<sigma> x" by blast+    
    show ?case 
      by (rule exI[of _ "D\<langle>r' \<cdot> \<tau>\<rangle>"], unfold l r D s', intro conjI, simp, rule term_subst_eq, rule \<tau>[symmetric], insert rule IH(1), unfold cover_def, auto)
  qed
qed


declare size_simps[simp]
fun height_ge where "height_ge (Var x) _ = True"
 | "height_ge (Fun f ts) (Fun g ss) = (height f \<ge> height g \<and> (\<forall> i < length ts. height_ge (ts ! i) (ss ! i)))"
declare size_simps[simp del]

(* definition 12 *)
definition height_gt where "height_gt s t \<equiv> base_term s = base_term t \<and> height_ge s t"

lemma height_ge_subst: assumes "base_term l = base_term r"
  shows "height_ge (l \<cdot> \<sigma>) (r \<cdot> \<tau>) = (height_ge l r \<and> (\<forall> x \<in> vars_term l. height_ge (get_subst \<sigma> x) (get_subst \<tau> x)))"
  using assms
proof (induct l arbitrary: r)
  case (Var x)
  hence r: "r = Var x" by (cases r, auto)
  thus ?case  by auto
next
  case (Fun f ls)
  from Fun(2)
  obtain g rs where r: "r = Fun g rs" by (cases r, auto)
  note Fun = Fun[unfolded r]
  from map_eq_imp_length_eq[OF Fun(2)[simplified, THEN conjunct2]]
  have len: "length ls = length rs" .
  let ?geb = "\<lambda> i. (height_ge (ls ! i \<cdot> \<sigma>) (rs ! i \<cdot> \<tau>))" 
  let ?ge = "\<lambda> i. (height_ge (ls ! i) (rs ! i))" 
  let ?ges = "\<lambda> i. (\<forall>x\<in>vars_term (ls ! i). height_ge (get_subst \<sigma> x) (get_subst \<tau> x))"
  let ?l = "(\<forall>i<length rs. ?geb i)"
  let ?r = " (\<forall>i<length rs. ?ge i) \<and>
    (\<forall>a\<in>set ls. \<forall>x\<in>vars_term a. height_ge (get_subst \<sigma> x) (get_subst \<tau> x))"
  let ?r' = " (\<forall>i<length rs. ?ge i) \<and>
    (\<forall>i < length ls. ?ges i)"
  {
    fix i
    assume i: "i < length rs"
    hence i': "i < length ls" unfolding len by auto
    hence mem: "ls ! i \<in> set ls" by auto
    have base: "base_term (ls ! i) = base_term (rs ! i)" 
      by (rule map_nth_conv[THEN spec, THEN mp, OF _ i'], insert Fun(2), auto)
    have "?geb i = (?ge i \<and> ?ges i)"
      by (rule Fun(1)[OF mem base])
  }
  hence "?l = ?r'" unfolding len by blast
  also have "?r = ?r'" unfolding all_set_conv_all_nth  len by auto
  finally show ?case unfolding r
    by (simp, simp add: len)
qed


(* lemma 13 *)
lemma height_gt_mono_single: assumes ll: "left_linear_trs R"
  and wf: "\<And> l r. (l,r) \<in> R \<Longrightarrow> vars_term r \<subseteq> vars_term l"
  and step: "(s1,t1) \<in> rstep (cover ff R)"
  and ge: "height_gt s2 s1"
  shows "\<exists> t2. (s2,t2) \<in> rstep (cover ff R) \<and> height_gt t2 t1"
proof -
  let ?R = "cover ff R"
  from ge have base:"base_term s1 = base_term s2" unfolding height_gt_def by simp  
  from step obtain D \<tau> l' r' where s1: "s1 = D\<langle>l' \<cdot> \<tau>\<rangle>" 
    and t1: "t1 = D\<langle>r' \<cdot> \<tau>\<rangle>"
    and rule: "(l',r') \<in> ?R"
    by (induct, auto)
  let ?C = "map_ctxt base D"
  let ?\<sigma> = "map_subst base \<tau>"
  from rule obtain l r where lr: "(l,r) \<in> R" and l'r': "(l',r') \<in> cover ff {(l,r)}"
    unfolding cover_def by auto
  from cover6a[OF l'r'] have l: "l = base_term l'" and r: "r = base_term r'"
    by auto
  have bs1: "base_term s1 = ?C\<langle>l \<cdot> ?\<sigma>\<rangle>" unfolding s1 l by simp
  from ll lr have lin: "linear_term l" unfolding left_linear_trs_def by auto
  from enrichment_cover_extended[OF lin wf[OF lr] bs1[unfolded base]]  
  obtain l'' r'' \<tau>'' D'' where rule'': "(l'',r'') \<in> cover ff {(l,r)}"
    and s2: "s2 = D''\<langle>l'' \<cdot> \<tau>''\<rangle>" and D2: "?C = map_ctxt base D''"
    and l2: "base_term l'' = l" and r2: "base_term r'' = r"
    and \<tau>2: "\<And> x. x \<in> vars_term r \<Longrightarrow> get_subst (map_subst base \<tau>'') x = get_subst (map_subst base \<tau>) x" by blast+    
  from l2 l have bl: "base_term l'' = base_term l'" by simp
  from r2 r have br: "base_term r'' = base_term r'" by simp
  show ?thesis unfolding s2 t1
  proof (rule exI[of _ "D''\<langle>r'' \<cdot> \<tau>''\<rangle>"], rule conjI)
    show "(D''\<langle>l'' \<cdot> \<tau>''\<rangle>,D''\<langle>r'' \<cdot> \<tau>''\<rangle>) \<in> rstep ?R" using rule'' lr 
      unfolding cover_def by auto
  next
    have beq: "base_term D''\<langle>r'' \<cdot> \<tau>''\<rangle> = base_term D\<langle>r' \<cdot> \<tau>\<rangle>"
      by (simp add: D2 r2 r, rule term_subst_eq, rule \<tau>2, unfold r, simp)     
    show "height_gt D''\<langle>r'' \<cdot> \<tau>''\<rangle> D\<langle>r' \<cdot> \<tau>\<rangle>" unfolding height_gt_def
    proof (rule conjI[OF beq])
      obtain l1 l2 r1 r2 where r1: "r'' \<cdot> \<tau>'' = r1" and r2: "r' \<cdot> \<tau> = r2" 
        and l1: "l'' \<cdot> \<tau>'' = l1" and l2: "l' \<cdot> \<tau> = l2" by auto
      show "height_ge D''\<langle>r'' \<cdot> \<tau>''\<rangle> D\<langle>r' \<cdot> \<tau>\<rangle>"  
        using ge[unfolded height_gt_def s1 s2, THEN conjunct2] D2[symmetric]
        unfolding l1 l2 r1 r2
      proof (induct D'' arbitrary: D)
        case (More f bef' C' aft' D)
        from More(3) obtain g bef C aft where D: "D = More g bef C aft"
          by (cases D, force+)
        note More = More[unfolded this]
        note eq = More(3)[simplified, THEN conjunct2] 
        from eq have "map_ctxt base C' = map_ctxt base C" by simp
        note ind = More(1)[OF _ this]
        from eq[THEN conjunct1] have lenb:"length bef' = length bef" by (rule map_eq_imp_length_eq)
        from More(2) have gf: "height g \<le> height f" by simp
        show ?case unfolding D
        proof (simp add: gf, intro allI impI)
          fix i
          let ?n = "Suc (length bef + length aft)"
          let ?j = "length bef"
          let ?k = "length bef'"
          assume i: "i < Suc (?k + length aft')"
          from More(2)[simplified] i have ge:
            "height_ge ((bef' @ C'\<langle>l1\<rangle> # aft') ! i) ((bef @ C\<langle>l2\<rangle> # aft) ! i)" 
            (is "height_ge ?bl1 ?bl2") by auto
          show "height_ge ((bef' @ C'\<langle>r1\<rangle> # aft') ! i) ((bef @ C\<langle>r2\<rangle> # aft) ! i)"            (is "height_ge ?br1 ?br2") 
          proof (cases "i < ?j")
            case True
            with lenb have "i < ?k" by auto
            with True have "?bl1 = bef' ! i" and "?bl2 = bef ! i" and r2: "?br2 = bef ! i" and r1: "?br1 = bef' ! i" unfolding nth_append by auto
            from ge[unfolded this] show ?thesis unfolding r1 r2 .
          next
            case False            
            note j = this
            hence k: "\<not> i < ?k" unfolding lenb by simp
            show ?thesis 
            proof (cases "i = ?j")
              case True
              with lenb have k:"i = ?k" by auto
              from True have l2: "?bl2 = C\<langle>l2\<rangle>" and r2: "?br2 = C\<langle>r2\<rangle>" unfolding nth_append by auto
              from k have l1: "?bl1 = C'\<langle>l1\<rangle>" and r1: "?br1 = C'\<langle>r1\<rangle>" unfolding nth_append by auto
              from ind[OF ge[unfolded l1 l2]]
              show ?thesis unfolding r1 r2 .
            next
              case False
              with j have "i - ?j > 0" by arith 
              then obtain ii where ij: "i - ?j = Suc ii" by (cases "i - ?j", auto)
              have bl2: "?bl2 = aft ! ii" and br2: "?br2 = aft ! ii" unfolding nth_append ij using j by auto
              from ij lenb have ik: "i - ?k = Suc ii" by auto
              have bl1: "?bl1 = aft' ! ii" and br1: "?br1 = aft' ! ii" unfolding nth_append ik using k by auto
              from ge[unfolded bl1 bl2]
              show ?thesis unfolding br1 br2 .
            qed
          qed
        qed
      next
        case Hole
        hence D: "D = Hole" by (cases D, auto)
        with Hole have "height_ge (l'' \<cdot> \<tau>'') (l' \<cdot> \<tau>)" using l1 l2 by simp
        from this[unfolded height_ge_subst[OF bl]]
        have ge: "height_ge l'' l'" and ges: "\<forall>x \<in> vars_term l''. height_ge (get_subst \<tau>'' x) (get_subst \<tau> x)" by auto
        let ?h = "\<lambda> l'. Suc (min_list (map height (sym_collect (\<lambda>t'. ff (l, r) (base_term t')) l')))"
        from rule'' obtain h'' where r'': "r'' = lift_term h'' r" and h'': "h'' = ?h l''" unfolding cover_def 
by blast
        from l'r' obtain h' where r': "r' = lift_term h' r" and h': "h' = ?h l'" unfolding cover_def by blast
        obtain gg where gg: " (\<lambda>t'. ff (l, r) (base_term t')) = gg" by auto
        let ?ml = "\<lambda> l'. min_list (map height (sym_collect gg l'))"
        let ?s = "sym_collect gg"
        let ?l = "\<lambda> l'. length (?s l')"
        let ?h = "height"
        from ge have "(\<forall> i < ?l l''. ?h (?s l' ! i) \<le> ?h (?s l'' ! i)) \<and> ?l l'' = ?l l'"
          using bl          
        proof (induct l'' arbitrary: l')
          case (Var x r)
          from Var(2) have r: "r = Var x" by (cases r, auto)
          thus ?case by auto
        next
          case (Fun f ls r)
          from Fun(3) obtain g rs where r: "r = Fun g rs" by (cases r, auto)
          note Fun = Fun[unfolded r]
          from Fun(3) gg have gg: "gg (Fun f ls) = gg (Fun g rs)" by auto
          from Fun(3) map_eq_imp_length_eq[of base_term ls base_term rs]
          have len: "length ls = length rs" by simp
          {
            fix i
            assume i: "i < length ls"
            hence mem: "ls ! i \<in> set ls" and i': "i < length rs" using len by auto            
            have base: "base_term (ls ! i) = base_term (rs ! i)" 
              by (rule map_nth_conv[THEN spec, THEN mp, OF _ i], insert Fun(3), auto)
            from Fun(2) i i' have "height_ge (ls ! i) (rs ! i)" by simp
            note Fun(1)[OF mem this base]
          } note ind = this
          let ?ls = "map height (concat (map (sym_collect gg) ls))"
          let ?rs = "map height (concat (map (sym_collect gg) rs))"
          let ?lls = "concat (map (\<lambda>x. map height (sym_collect gg x)) ls)"
          let ?rrs = "concat (map (\<lambda>x. map height (sym_collect gg x)) rs)"
          have ls: "?ls = ?lls"  unfolding map_concat map_map o_def ..
          have rs: "?rs = ?rrs"  unfolding map_concat map_map o_def ..
          obtain lls rrs where rrs: "rrs = ?rrs" and lls: "lls = ?rrs" by auto
          have lenc:"length ?ls = length ?rs" unfolding ls rs
            by (rule eq_length_concat_nth, simp add: len, simp add: len ind)
          hence lenci: "length (concat (map ?s ls)) = length (concat (map ?s rs))" by simp
          {
            fix i
            assume "i < length (map ?s ls)"
            hence i: "i < length ls" and i': "i < length rs" using len by auto
            have "length (map ?s ls ! i) = length (map ?s rs ! i)" 
              using ind[OF i] unfolding nth_map[OF i] nth_map[OF i'] by simp
          } note le = this
          have ge: "\<forall>i<length (concat (map ?s ls)).
            ?h (concat (map ?s rs) ! i) \<le> ?h (concat (map ?s ls) ! i)"
          proof (rule concat_all_nth, simp add: len, rule le)
            fix i j
            assume i: "i < length (map ?s ls)" and j: "j < length (map ?s ls ! i)"
            with le[OF i] have i: "i < length ls" and i': "i < length rs" and j': "j < length (map ?s rs ! i)" using len by auto
            show "height (map ?s rs ! i ! j) \<le> height (map ?s ls ! i ! j)"
              unfolding nth_map[OF i] nth_map[OF j] nth_map[OF i'] nth_map[OF j']
              by (rule ind[OF i, THEN conjunct1, THEN spec, THEN mp], insert j, unfold nth_map[OF i], simp)
          qed          
          show ?case
          proof (cases "gg (Fun g rs)")
            case False
            show ?thesis 
              by (unfold r sym_collect.simps gg, simp add: False lenci ge) 
          next
            case True
            show ?thesis 
            proof (unfold r sym_collect.simps gg, simp add: True,
              rule conjI[OF _ lenci], unfold all_Suc, simp add: ge)
              from True Fun(2)
              show "height g \<le> height f"
                by (simp)
            qed
          qed
        qed
        note main = this
        have ge: "h' \<le> h''" unfolding h' h'' gg 
          by (simp, rule min_list_nth, insert main, auto)
        hence ge: "height_ge r'' r'" unfolding r' r''
          by (induct r, auto)
        have "vars_term r'' = vars_term r" unfolding r br[symmetric] by simp
        also have "... \<subseteq> vars_term l" using wf[OF lr] .
        also have "... = vars_term l''" unfolding l bl[symmetric] by simp
        finally have ges: "\<forall>x \<in> vars_term r''. height_ge (get_subst \<tau>'' x) (get_subst \<tau> x)" using ges by auto
        have "height_ge (r'' \<cdot> \<tau>'') (r' \<cdot> \<tau>)" 
          unfolding height_ge_subst[OF br] using ge ges by auto
        thus ?case using D r1 r2 by simp
      qed
    qed
  qed
qed
              

lemma height_gt_mono: assumes ll: "left_linear_trs R"
  and wf: "\<And> l r. (l,r) \<in> R \<Longrightarrow> vars_term r \<subseteq> vars_term l"
  and steps: "(s1,t1) \<in> (rstep (cover ff R))^*"
  and ge: "height_gt s2 s1"
  shows "\<exists> t2. (s2,t2) \<in> (rstep (cover ff R))^* \<and> height_gt t2 t1"
  using steps ge
proof (induct arbitrary: s2)
  case (base s2)
  show ?case
    by (rule exI[of _ s2], insert base, unfold height_gt_def, auto)
next
  case (step t1 u1 s2)
  let ?R = "rstep (cover ff R)"
  from step(3)[OF step(4)] obtain t2 where steps: "(s2,t2) \<in> ?R^*" and ht: "height_gt t2 t1" by auto
  from height_gt_mono_single[OF ll wf step(2) ht] obtain u2 where step: "(t2,u2) \<in> ?R"
    and hu: "height_gt u2 u1" by auto
  show ?case
    by (rule exI[of _ u2], simp add: hu, insert steps step, auto)
qed


fun roof :: "('f,'v)rule \<Rightarrow> ('f,'v) term \<Rightarrow> bool"
  where [code del]: "roof (l,r) = (\<lambda> t. vars_term t \<subseteq> vars_term l)"

lemma [code]:
  "roof (l,r) = (let xs = vars_term_list l in (\<lambda> t. list_all (\<lambda> x. x \<in> set xs) (vars_term_list t)))" unfolding Let_def list_all_iff roof.simps vars_term_list_sound
  by (force simp: mem_def)


end
